--[[
Name: LibSpreadsheet-1.0
Revision: $Rev: 11 $
Author(s): Humbedooh
Description: Spreadsheet stuff
Dependencies: LibStub AceGUI-3.0
License: GPL v3 or later.
]]
-----------------------------------------------------------------------

local SPREADSHEET = "LibSpreadsheet-1.0"
local SPREADSHEET_MINOR = tonumber(("$Rev: 11 $"):match("(%d+)"))
if not LibStub then error(SPREADSHEET .. " requires LibStub.") end
local lib = LibStub:NewLibrary(SPREADSHEET, SPREADSHEET_MINOR)
if not lib then return end

lib.disabled = lib.disabled or nil;
lib.objects = lib.objects or {};
lib.sheets = lib.sheets or {};
lib.str = UIParent:CreateFontString();
lib.cols = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z', '?'};

function lib:AddSheet(title)
	local obj = self;
	if ( type(title) ~= "string" ) then error(("%s: Error in setting sheet title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(obj.sheets) do
		if ( v.title == title ) then error(("%s: Sheet \"%s\" already exists in book \"%s\"!"):format(SPREADSHEET, title, obj.title)); end
	end
	local sheet = {
        parent = obj,
		type = 'sheet',
		title = title,
		columns = {},
        lastColumn = 0,
		rows = {},
		rowHeight = 0,
		lastRow = 0,
		lastColumn = 0,
        currentRow = nil,
        currentColumn = nil,
		frame = nil,
		AddColumn = lib.AddColumn,
		SetColumn = lib.SetColumn,
		DeleteColumn = lib.DeleteColumn,	
        ExtractLocation = lib.ExtractLocation,
		AddRow = lib.AddRow,
		GoTo = lib.GoTo,
		Insert = lib.Insert,
		Delete = lib.Delete,
        isInteractive = false,
        StartSelect = lib.StartSelect,
        EndSelect = lib.EndSelect,
        SetVisibleSelection = lib.SetVisibleSelection,
        UpdateSelect = lib.UpdateSelect,
        SetInteractive = lib.SetInteractive,
        SetColorRange = lib.SetColorRange,
        SetColor = lib.SetColor,
        ClearCells = lib.ClearCells,
	}
	table.insert(obj.sheets, sheet);
	return obj:Select(sheet.title);
end

function lib:SetInteractive(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting interactive mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in editing sheet: Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    sheet.isInteractive = toggle;
end

function lib:DeleteSheet(title)
	local obj = self;
    local i = nil;
	if ( type(title) ~= "string" ) then error(("%s: Error in deleting sheet; Invalid sheet name - expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(obj.sheets) do
		if ( v.title == title ) then
            if ( obj.selectedSheet == v ) then
                obj.selectedSheet = nil;
            end
			i = k;
			break;
		end		
	end
    if ( i~= nil ) then
        table.remove(obj.sheets, i);
    end
end

function lib:Select(sheet)
	local obj = self;
	for k, v in pairs(obj.sheets) do
		if ( v.title == sheet ) then
            obj.selectedSheet = v;
            obj.innerScrollbar:SetValue(0);
			return v;
		end
	end
	return nil;
end

function lib:ClearCells()
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while clearing sheet, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    table.wipe(sheet.rows);
end
function lib:AddColumn(...)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while adding column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
	for i = 1, select("#", ...) do
		local column = select(i, ...);
        if ( type(column) == "string" ) then
            lib.str:SetFontObject("LibSpreadsheet_StringFont");
            lib.str:SetText(column);
            local w = lib.str:GetStringWidth() + 5;
			local h = lib.str:GetStringHeight() + 2;
            if ( h > obj.rowHeight ) then obj.rowHeight = h; end
            table.insert(obj.columns, {type="string", width=w, name=column, justify="LEFT", color = {1,1,1}});
            obj.lastColumn = obj.lastColumn + 1;
        elseif ( type(column) == "table" ) then
            local col = {name="", type="string"};
            if ( column.name ~= nil ) then col.name = column.name; end
            if ( column.type == "string" or column.type == "number" or column.type == "boolean" ) then
                col.type = column.type;
            end
            if ( col.type == "number" ) then
                lib.str:SetFontObject("LibSpreadsheet_NumberFont");
                col.justify = "RIGHT";
            else
                lib.str:SetFontObject("LibSpreadsheet_StringFont");
                col.justify = "LEFT";
            end
            col.minval = column.minval or nil;
            col.maxval = column.maxval or nil;
            col.interactive = column.interactive or false;
            col.onClick = column.onClick or nil;
            col.color = column.color or {1,1,1};
            col.suffix = column.suffix or "";
            col.justify = column.justify or col.justify;
            lib.str:SetText(col.name .. col.suffix .. ": ");
            local w = lib.str:GetStringWidth() + 5;
			local h = lib.str:GetStringHeight() + 2;
            if ( h > obj.rowHeight ) then obj.rowHeight = h; end
            col.width = w;
            if ( column.decimals ~= nil ) then
                col.decimals = column.decimals;
            end
            obj.lastColumn = obj.lastColumn + 1;
            table.insert(obj.columns, col);
        else
            error(("%s: Error while adding column, expected string or table, got %s!"):format(SPREADSHEET, column));
        end
	end
end

function lib:SetColorRange(column, ...)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while setting color range, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    if ( type(column) == "string" ) then
        column = tonumber(column, 36) - 9;
    end
    if ( sheet.columns[column] == nil ) then error(("%s: Error while setting color range for column %s: No such column!"):format(SPREADSHEET, lib.cols[column])); end
    sheet.columns[column].colorRange = {};
    for i = 1, select("#", ...) do
		local color = select(i, ...);
        table.insert(sheet.columns[column].colorRange, color);
    end
end


function lib:SetColumn(column, properties)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while setting column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local c = column
    if ( type(column) == "string" ) then
        c = tonumber(column, 36) - 9;
    end
    if ( sheet.columns[c] == nil ) then error(("%s: Error while setting column %s: No such column!"):format(SPREADSHEET, lib.cols[c])); end
    if ( type(properties) == "string" ) then
        lib.str:SetText(properties);
        local w = lib.str:GetStringWidth() + 5;
        local h = lib.str:GetStringHeight() + 2;
        if ( h > sheet.rowHeight ) then sheet.rowHeight = h; end
        sheet.columns[c].width = w;
        sheet.columns[c].name=properties;
    elseif ( type(properties) == "table" ) then
        if ( properties.name ~= nil ) then sheet.columns[c].name = properties.name; end
        if ( properties.type == "string" or properties.type == "number" or properties.type == "boolean" ) then
            sheet.columns[c].type = properties.type;
        end
        sheet.columns[c].justify = properties.justify or sheet.columns[c].justify;
        if ( sheet.columns[c].type == "number" ) then
            lib.str:SetFontObject("LibSpreadsheet_NumberFont");
        else
            lib.str:SetFontObject("LibSpreadsheet_StringFont");
        end
        sheet.columns[c].suffix = properties.suffix or sheet.columns[c].suffix;
        lib.str:SetText(sheet.columns[c].name .. sheet.columns[c].suffix .. ": ");
        sheet.columns[c].color = properties.color or {1,1,1};
        sheet.columns[c].minval = properties.minval or sheet.columns[c].minval;
        sheet.columns[c].maxval = properties.maxval or sheet.columns[c].maxval;
        local w = lib.str:GetStringWidth() + 5;
        local h = lib.str:GetStringHeight() + 2;
        if ( h > sheet.rowHeight ) then sheet.rowHeight = h; end
        sheet.columns[c].width = w;
        if ( properties.decimals ~= nil ) then
            sheet.columns[c].decimals = properties.decimals;
        end
    else
        error(("%s: Error while setting column properties, expected string or table, got %s!"):format(SPREADSHEET, type(properties)));
    end
end

function deleteColumn(column)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while deleting column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local c = nil;
    if ( type(column) == "string" ) then
        c = lib.cols[string.upper(column)];
    else
        c = tonumber(column);
        if ( c > 26 ) then c = 26; end
    end
    if ( sheet.columns[c] == nil ) then error(("%s: Error while deleting column %s: No such column!"):format(SPREADSHEET, lib.cols[c])); end
    table.remove(sheet.columns, c);
    for k, v in pairs(sheet.rows) do
        table.remove(v, c);
    end
end

function lib:AddRow(...)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while adding row, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local row = {};
    for i = 1, select("#", ...) do
		local data = select(i, ...);
        local cell = {};
        if ( type(data) ~= "table" ) then
            cell.data = data;
            cell.color = nil;
        else
            cell.data = data.data;
            cell.color = data.color or nil;
            cell.id = data.id or nil;
        end
        if ( sheet.parent.strict == true ) then
            local t =  sheet.columns[i].type;
            if ( t == "string" and type(cell.data) ~= "string" ) then error(("%s: Error while adding cell in column %s, expected string, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell.data))); end
            if ( t == "boolean" and type(cell.data) ~= "boolean" ) then error(("%s: Error while adding cell in column %s, expected boolean, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell.data))); end
            if ( ( t == "number" ) and type(cell.data)  ~= "number" ) then error(("%s: Error while adding cell in column %s, expected number, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell.data))); end
        end
        if ( sheet.columns[i] ~= nil ) then
            if ( sheet.columns[i].type == "string" ) then
                lib.str:SetFontObject("LibSpreadsheet_StringFont");
                if ( type(cell.data) == "boolean" ) then
                    if ( cell.data == true ) then cell.data = "true";
                    else cell.data = "false";
                    end
                end
                lib.str:SetText(cell.data);
                local w = lib.str:GetStringWidth() + 5;
                if ( w > sheet.columns[i].width ) then sheet.columns[i].width= w; end
            elseif ( sheet.columns[i].type == "number" ) then
                local s = ("%%.%uf%%s"):format(sheet.columns[i].decimals or 0);
                lib.str:SetFontObject("LibSpreadsheet_NumberFont");
                lib.str:SetText(string.format(s, cell.data, (sheet.columns[i].suffix or "")));
                local w = lib.str:GetStringWidth() + 5;
                if ( w > sheet.columns[i].width ) then sheet.columns[i].width= w; end
            end
        end
        table.insert(row, cell);
    end
    table.insert(sheet.rows, row);
end

function lib:ExtractLocation(str)
    local t = (self.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while looking up column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = self;
    local col = nil;
    local row = nil;
    if ( type(str) == "string" ) then
        local c, r = (str):match("(%D+)(%d*)");
        if ( r ~= nil ) then
            row = r;
        end
        if ( string.len(c) == 1 ) then
            col = tonumber(c, 36) - 9;
        else
            for k, column in pairs(sheet.columns) do
                if ( column.name == c ) then
                    col = k;
                    break;
                end
            end
        end
    elseif ( type(column) == "number" ) then
        if ( column > 26 ) then column = math.fmod(column, 26); end
        if ( column < 1 ) then 
            while ( column < 1 ) do
                column = column + sheet.lastColumn;
            end
        end
    end
    return col, tonumber(row);
end

function lib:SetVisibleSelection(fromCol, fromRow, toCol, toRow)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in GoTo(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    if ( type(fromCol) == "number" ) then
        sheet.selectStartColumn = fromCol;
        sheet.selectStartRow = fromRow;
    elseif ( type(fromCol) == "string" ) then
        local c, r = sheet:ExtractLocation(fromCol);
        if ( type(r) == "number" ) then
            sheet.selectStartColumn = c;
            sheet.selectStartRow = r;
            toRow = toCol;
            toCol = fromRow;
        else
            sheet.selectStartColumn = c;
            sheet.selectStartRow = fromRow;
        end
    end
    if ( type(toCol) == "number" ) then
        sheet.selectEndColumn = toCol;
        sheet.selectEndRow = toRow;
    elseif ( type(toCol) == "string" ) then
        local c, r = sheet:ExtractLocation(toCol);
        if ( type(r) == "number" ) then
            sheet.selectEndColumn = c;
            sheet.selectEndRow = r;
        else
            sheet.selectEndColumn = c;
            sheet.selectEndRow = toRow;
        end
    end
    sheet.selecting = true;
    sheet:UpdateSelect(nil);
    sheet.selecting = false;
end

function lib:GoTo(column, row) -- takes arguments like: GoTo("B"), GoTo("B1"), GoTo("B", 3), GoTo(2,5), GoTo("Name", 3), GoTo("Level5");
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in GoTo(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
	local c, r = sheet:ExtractLocation(column);
    if ( type(r) == "number" ) then
        row = r;
    end
    if ( sheet.columns[column] == nil ) then error(("%s: Error in GoTo(): No such column '%s'!"):format(SPREADSHEET, lib.cols[column or 27])); end
    if ( sheet.rows[row] == nil ) then error(("%s: Error in GoTo(): No such row '%u'!"):format(SPREADSHEET, row)); end
    sheet.currentColumn = column;
    sheet.currentRow = row;    
end

function lib:Delete(fromCol, fromRow, toCol, toRow)
	
end

function lib:Insert(...)
	
end

function lib:StartSelect(caller, event)
    local sheet = self;
    sheet.selecting = true;
    sheet.selectStartColumn = caller.column;
    sheet.selectStartRow = caller.row;
    if ( event == "RightButton" ) then
        sheet.selectStartColumn = nil;
        sheet.selectStartRow = nil;
        sheet.selectEndColumn = nil;
        sheet.selectEndRow = nil;
    end
    sheet:UpdateSelect(caller);
end

function lib:EndSelect(caller)
    local sheet = self;
    sheet.selectEndColumn = caller.column;
    sheet.selectEndRow = caller.row;
    sheet.selecting = false;
    sheet:UpdateSelect(caller);
end

function lib:UpdateSelect(caller)
    local sheet = self;
    if ( sheet.selecting or false == true ) then
        if ( caller ~= nil ) then
            sheet.selectEndColumn = caller.column;
            sheet.selectEndRow = caller.row;
        end
        for x = 1, table.maxn(sheet.rows) do
            for y = 1, sheet.lastColumn do
                sheet.rows[x][y].frame.tex:SetTexture(0,0,0,0);
            end
        end
        if ( sheet.selectEndRow == nil ) then sheet.selectEndRow  = sheet.selectStartRow ; end
        if ( sheet.selectStartRow ~= nil ) then
            for x = math.min(sheet.selectStartRow, sheet.selectEndRow), math.max(sheet.selectStartRow, sheet.selectEndRow) do
                for y = math.min(sheet.selectStartColumn, sheet.selectEndColumn), math.max(sheet.selectStartColumn, sheet.selectEndColumn) do
                    sheet.rows[x][y].frame.tex:SetTexture(0.3,0.2,1,0.25);
                end
            end
        end
    end
end


function lib:Build()
    local obj = self;
    local sheet = obj.selectedSheet;
    if ( obj.innerFrame ~= nil ) then
        obj.innerFrame:Hide();
        obj.innerFrame:SetParent(nil);
    end
    if ( obj.header ~= nil ) then
        obj.header:Hide();
        obj.header:SetParent(nil);
        obj.header = nil;
    end
    obj.innerScrollframe:SetPoint("TOPLEFT", obj.outerScrollframe, "TOPLEFT", 0, -sheet.rowHeight);
    sheet.frame = CreateFrame("Frame", nil, obj.innerScrollframe);
    sheet.frame:SetAllPoints();
    obj.innerFrame = sheet.frame;
    sheet.frame:SetWidth(obj.innerScrollframe:GetWidth());
    obj.innerScrollframe:SetScrollChild(sheet.frame);
    obj.innerScrollbar:Hide();
    sheet.frame:Show();
    obj.header = CreateFrame("Frame", nil, obj.frame);
    obj.header:SetPoint("TOPLEFT", obj.outerScrollframe, "TOPLEFT", 0, 0);
    obj.header:SetPoint("BOTTOMRIGHT", obj.outerScrollframe, "TOPRIGHT", 0, -( sheet.rowHeight+10));
    obj.header:Show();
    
    local offset = 0;
    for key, cell in pairs (sheet.columns) do
        local cellFrame = CreateFrame("Frame", nil, obj.header);
        cellFrame.label = cellFrame:CreateFontString(nil, "DIALOG", LibSpreadsheet_StringFont);
        cellFrame.label:SetAllPoints(cellFrame);
        cellFrame:Show();
        if ( cell.click ~= nil ) then
            --label:SetScript("OnClick", cell.click);
        end
        cellFrame.label:SetFontObject(LibSpreadsheet_StringFont);
        cellFrame:SetWidth(cell.width);
        cellFrame:SetHeight(sheet.rowHeight);
        cellFrame.label:SetText((cell.name or "??")..":");
        cellFrame.label:SetJustifyH(cell.justify);
        cellFrame:SetPoint("TOPLEFT", obj.header, "TOPLEFT", offset, 1);
        offset = offset + cell.width + (sheet.rowHeight/2);        
    end
    obj.header:SetPoint("TOP", obj.frame, "TOP", 0, 0);
    
    
    local offsetY = 5;
    for rkey, rowitem in pairs (sheet.rows) do
        local offsetX = 0;
        local row = CreateFrame("Frame", nil, sheet.frame);
        row:SetPoint("TOPLEFT", sheet.frame, "TOPLEFT", 0, -offsetY);
        row:SetPoint("TOPRIGHT", sheet.frame, "TOPRIGHT", 0, -offsetY);
        row:SetHeight(sheet.rowHeight + 5);
        local i = 0;
        for key, cell in pairs(rowitem) do
            i = i + 1;
            local cellFrame = CreateFrame("Button", nil, row);
            cellFrame.label = cellFrame:CreateFontString(nil, "DIALOG", LibSpreadsheet_StringFont);
            cellFrame:SetPoint("TOPLEFT", row, "TOPLEFT", offsetX, 0);
            offsetX = offsetX + sheet.columns[i].width + (sheet.rowHeight/2);
            cellFrame.label:SetAllPoints(cellFrame);
            if ( cell.click ~= nil ) then
                --label:SetCallback("OnClick", cell.click);
            else
                cellFrame.tex = cellFrame:CreateTexture(nil, "OVERLAY");
                cellFrame.tex:SetAllPoints();
                cellFrame.tex:SetBlendMode("ADD");
                cellFrame.column = i;
                cellFrame.row = rkey;
                if ( sheet.isInteractive or sheet.columns[i].interactive ) then
                    cellFrame:SetScript("OnMouseDown", function(caller, event) sheet:StartSelect(cellFrame, event); end);
                    cellFrame:SetScript("OnMouseUp", function(caller, event) sheet:EndSelect(cellFrame); end);
                    cellFrame:SetScript("OnEnter", function(caller, event) sheet:UpdateSelect(cellFrame); end);
                    cellFrame:SetScript("OnLeave", function(caller, event) sheet:UpdateSelect(cellFrame); end);
                end
                if ( sheet.columns[i].onClick ~= nil ) then
                    local func = sheet.columns[i].onClick;
                    cellFrame:SetScript("OnClick", function(caller, event) func(cell.id, cell.data, cellFrame); end);

                end
                cell.frame = cellFrame;
            end
            cellFrame.label:SetFontObject(LibSpreadsheet_StringFont);
            cellFrame:SetWidth(sheet.columns[i].width);
            cellFrame:SetHeight(sheet.rowHeight);
            cellFrame.label:SetJustifyH(sheet.columns[i].justify);
            cellFrame.label:SetTextColor(sheet.columns[i].color[1], sheet.columns[i].color[2], sheet.columns[i].color[3]);
            if ( cell.color ~= nil ) then
                cellFrame.label:SetTextColor(cell.color[1] or cell.color.r, cell.color[2] or cell.color.g, cell.color[3] or cell.color.b);
            end
            if ( sheet.columns[i].type == "number" ) then
                cellFrame.label:SetFontObject(LibSpreadsheet_NumberFont);
                if ( sheet.columns[i].decimals ~= nil ) then
                    local s = ("%%.%uf%%s"):format(sheet.columns[i].decimals);
                    cellFrame.label:SetText(string.format(s, cell.data, (sheet.columns[i].suffix or "")));
                else
                    cellFrame.label:SetText(string.format("%i%s", cell.data, (sheet.columns[i].suffix or "")));
                end
                if ( sheet.columns[i].colorRange ~= nil ) then
                    if ( sheet.columns[i].minval ~= nil and sheet.columns[i].maxval ~= nil ) then
                        local colors = table.maxn(sheet.columns[i].colorRange);
                        local where = ( ( (cell.data - sheet.columns[i].minval)  / (sheet.columns[i].maxval - sheet.columns[i].minval)) * (colors-1) ) + 1;
                        local p, n = math.floor(where), math.ceil(where);
                        local r = (where - p) * 100;
                        if (  sheet.columns[i].colorRange[p] == nil ) then
                            where = ((colors-1)/2) + 1
                            p, n = math.floor(where), math.ceil(where);
                            r = 0;
                        end
                        local color = { 
                            (  ( sheet.columns[i].colorRange[p][1] * (100-r) ) + ( sheet.columns[i].colorRange[n][1] * (r) ) ) / 100,
                            (  ( sheet.columns[i].colorRange[p][2] * (100-r) ) + ( sheet.columns[i].colorRange[n][2] * (r) ) ) / 100,
                            (  ( sheet.columns[i].colorRange[p][3] * (100-r) ) + ( sheet.columns[i].colorRange[n][3] * (r) ) ) / 100
                            };
                        cellFrame.label:SetTextColor(color[1], color[2], color[3]);
                    end
                end
            elseif ( sheet.columns[i].type == "string" ) then
                cellFrame.label:SetText(cell.data);
            elseif ( sheet.columns[i].type == "boolean" ) then
                if ( cell.data == true ) then cellFrame.label:SetText("true");
                else cellFrame.label:SetText("false"); end
            end
        end
        offsetY = offsetY + sheet.rowHeight + 2;
    end
    obj.innerScrollbar:SetMinMaxValues(0, offsetY);
    obj.innerScrollbar:SetValue(0);
    obj.innerFrame:ClearAllPoints()
    obj.innerFrame:SetPoint("TOPLEFT", 0, 0)
    obj.innerFrame:SetPoint("TOPRIGHT", 0, 0)
    obj.innerFrame:SetHeight(offsetY);
    obj.innerScrollframe:scrollFunc(0);
    obj.innerScrollframe:onMouseWheel(0);
    return obj.outerScrollframe;
end

function lib:StrictMode(toggle)
	if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting strict mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.strict = toggle;
end

function lib:Book(title)
	local obj = {};
	if ( type(title) ~= "string" ) then error(("%s: Error in creating book title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
	obj.sheets = {};
    obj.selectedSheet = nil;
	obj.type = 'book';
	obj.title = title;
    obj.strict = false;
	obj.Add = lib.AddSheet;
	obj.Delete = lib.DeleteSheet;
	obj.Select = lib.Select;
	obj.StrictMode = lib.StrictMode;
    obj.Build = lib.Build;
    obj.outerScrollframe = CreateFrame("ScrollFrame", nil, UIParent);
    obj.header = nil;
    obj.frame = CreateFrame("Frame", nil, obj.outerScrollframe);
    obj.frame:SetWidth(500);
    obj.frame:SetHeight(100);
    obj.frame:SetParent(UIParent);
    obj.frame:SetPoint("CENTER", UIParent, "CENTER", 0,0);
    obj.frame:SetFrameStrata("DIALOG")
    obj.frame:EnableMouse(true)
    obj.innerFrame = nil;
    obj.outerScrollframe:Hide();
    
    
	obj.outerScrollframe:SetAllPoints();
    obj.outerScrollframe:EnableMouse(true)
    
    obj.innerScrollframe = CreateFrame("ScrollFrame", nil, obj.outerScrollframe);
	obj.innerScrollframe:SetPoint("TOPLEFT");
	obj.innerScrollframe:SetPoint("BOTTOMRIGHT");
	obj.innerScrollframe:EnableMouseWheel(true)
    obj.innerScrollframe:EnableMouse(true)
    obj.outerScrollframe:SetScrollChild(obj.frame);
    obj.frame:SetAllPoints();
    obj.innerScrollframe:SetAllPoints();
 
    obj.innerScrollbar = CreateFrame("Slider", nil, obj.innerScrollframe, "UIPanelScrollBarTemplate");
	obj.innerScrollbar:SetPoint("TOPLEFT", obj.innerScrollframe, "TOPRIGHT", 4, -16)
	obj.innerScrollbar:SetPoint("BOTTOMLEFT", obj.innerScrollframe, "BOTTOMRIGHT", 4, 16)
	obj.innerScrollbar:SetMinMaxValues(0, 1000)
	obj.innerScrollbar:SetValueStep(1)
	obj.innerScrollbar:SetValue(0)
	obj.innerScrollbar:SetWidth(16)
    obj.innerScrollframe.scrollvalue = 0;
	--obj.innerScrollbar:Hide()
	obj.innerScrollbar:SetScript("OnValueChanged", function(a,b) obj.innerScrollframe:scrollFunc(b); end );
        
    
    obj.innerScrollframe.onMouseWheel = function(self, value)
		local height, viewheight = obj.innerScrollframe:GetHeight(), obj.innerFrame:GetHeight()
		
		if height > viewheight then
			obj.innerScrollbar:Hide()
		else
            obj.innerScrollbar:Show()
			local diff = height - viewheight
			local delta = 1
			if (value or 0) < 0 then
				delta = -1
			end
			obj.innerScrollbar:SetValue(min(max(obj.innerScrollframe.scrollvalue + delta*(1000/(diff/45)),0), 1000))
		end
	end;
	obj.innerScrollframe:SetScript("OnMouseWheel", obj.innerScrollframe.onMouseWheel);
    obj.outerScrollframe:SetScript("OnShow", function(self) obj.innerScrollframe:onMouseWheel(0); end);
    obj.innerScrollframe.scrollFunc = function(self, value)
		local viewheight = obj.innerScrollframe:GetHeight()
		local height = obj.innerFrame:GetHeight()
		local offset

		if viewheight > height then
			offset = 0
            obj.innerScrollbar:Hide()
		else
            obj.innerScrollbar:SetMinMaxValues(0, height-viewheight)
            obj.innerScrollbar:Show()
			offset = value;
		end
		obj.innerFrame:ClearAllPoints()
		obj.innerFrame:SetPoint("TOPLEFT", 0, offset)
		obj.innerFrame:SetPoint("TOPRIGHT", 0, offset)
		
		obj.innerScrollframe.scrollvalue = value
    end;
    
	return obj;
end

function LibSpreadTest()
    local lss = LibStub("LibSpreadsheet-1.0", true);                    -- Load the lib
    local book = lss:Book("myBook");                                    -- Make a new book
    local sheet = book:Add("my stuff");                                 -- Add a sheet to the book
    book:StrictMode(true);
    sheet.isInteractive = true;
    sheet:AddColumn(
        "Name",
        "Class",
        {name="Level", type="number", minval=1, maxval=80},
        {name="Score", type="number", decimals=2, suffix="%"},
        {name="Uses LibSpreadsheet!", type="boolean", color={0,1,1}}
        );  -- Set some headers
    sheet:AddRow("Humbedooh", "Warrior", 80, 1.34, true);                     -- Add some data
    sheet:AddRow("Kitiania", "|cffFF4499Priest|r", 27, 82.5, false);                      -- Add another row
    sheet:AddRow("Wimzig", "Gnome....", 50, 21, true);                     -- Add some data
    sheet:AddRow("Humbedooh", "Warrior", 60, 1.34, true);                     -- Add some data
    sheet:AddRow("Kitiania", "|cffFF4499Priest|r", 80, 82.5, false);                      -- Add another row
    sheet:AddRow("Wimzig", "Gnome....", 1, 21, true);                     -- Add some data
    sheet:AddRow("Humbedooh", "Warrior", 40, 1.34, true);                     -- Add some data
    sheet:AddRow("Kitiania", "|cffFF4499Priest|r", 75, 82.5, false);                      -- Add another row
    sheet:AddRow("Wimzig", "Gnome....", 12, 21, true);                     -- Add some data
    sheet:SetColorRange("C", {1,0,0}, {1,1,0}, {0,1,0});
    
    local frame = book:Build();                                        -- Create the UI
    frame:Show();
    
end